require "./vector"

module Physics

  class Tick
    property :n, :time, :last
    def initialize(@n : Int64, @time : Time, @last : Tick? = nil)
    end

    def timelaps : Time::Span
      last = @last
      if last.nil?
        return Time::Span.new
      else
        return @time - last.time
      end
    end

    class Timer
      getter :tick

      def initialize
        @tick = Tick.new(n: 0, time: Time.utc, last: nil)
      end

      def timeskip!(seconds : Int = 0, milliseconds : Int = 0, nanoseconds : Int = 0)
        timeskip!(timespan: Time::Span.new(seconds: seconds, nanoseconds: nanoseconds + milliseconds * 1_000_000))
      end

      def timeskip!(timespan : Time::Span)
        @tick = Tick.new(n: @tick.n + 1, time: @tick.time + timespan, last: @tick)
        self
      end
    end
  end

  def self.accelerate_by_one_tick!(tick : Tick, speed : Vector(T, N), acceleration : Vector(T, N)) : Vector(T, N) forall T, N
    fraction = tick.timelaps.total_milliseconds / 1000.0
    speed.add!(acceleration * fraction)
  end

  def self.move_by_one_tick!(tick : Tick, position : Vector(T, N), speed : Vector(T, N), acceleration : Vector(T, N)) : Vector(T, N) forall T, N
    time_fraction = tick.timelaps.total_milliseconds / 1000.0

    # TODO optimise memory using mutation of intermediate values
    max_speed = speed + (acceleration * time_fraction)
    tick_average_speed = (speed + max_speed) * 0.5 * time_fraction

    position.add!(tick_average_speed)
    speed.add!(acceleration * time_fraction)
  end

end
