module UI
  class Event
    getter :type, :code, :alt, :control, :shift, :system
    @type : Type
    @code : String?
    @alt : String?
    @control : String?
    @shift : String?
    @system : String?

    {% if @top_level.has_constant?(:SF) %}
    def initialize(event : SF::Event)
      @type = SF_TO_TYPE[event.class]? || Type::Unknown
      @code = event.code
      @alt = event.alt
      @control = event.control
      @shift = event.shift
      @system = event.system
    end
    {% end %}

    enum Type
      KeyPressed
      KeyReleased
      MouseButtonPressed
      MouseButtonReleased
      Unknown
    end

    {% if @top_level.has_constant?(:SF) %}
    # can this be a named tuple instead ?
    SF_TO_TYPE = {
      SF::Event::KeyPressed => Type::KeyPressed,
      SF::Event::KeyReleased => Type::KeyReleased,
      SF::Event::MouseButtonPressed => Type::MouseButtonPressed,
      SF::Event::MouseButtonReleased => Type::MouseButtonReleased,
    }
    {% end %}
    # `Event::Handler` is able to accumulate callbacks that will be triggered every
    # time a event of a given type is received.
    class Handler
      # A callback takes one argument (itself) and returns nothing
      alias Callback = Callback, Event -> Nil

      # NOTE Event.class should return Class so it's not yet what I want
      def initialize
        @events_callbacks = Hash(Type, Array(Callback)).new
      end

      def event_callbacks(event : Type) : Array(Callback)
        @events_callbacks[event] ||= Array(Callback).new
        @events_callbacks[event]
      end

      def add(event_type : Type, &block : Callback, Event -> Nil)
        event_callbacks(event_type) << block
        block
      end

      def add(event_type : Type, block : Callback)
        event_callbacks(event_type) << block
        block
      end

      def remove(event_type : Type, &block : Callback, Event -> Nil)
        event_callbacks(event_type).delete(block)
      end

      def remove(event_type : Type, block : Callback)
        event_callbacks(event_type).delete(block)
      end

      def handle(event : Event)
        event_callbacks(event.type).each { |callback| callback.call(callback, event) }
      end

      {% if @top_level.has_constant?(:SF) %}
      def event_callbacks(event : SF::Event.class) : Array(Callback)
        event = SF_TO_TYPE[event]? || Type::Unknown
        event_callbacks(event)
      end

      def handle(event : SF::Event)
        handle(Event.new(event))
      end
      {% end %}
    end
  end
end
